/*
 * fontserver.c -- deal with finding a font server and
 * reading its config file
 */

#include <stdio.h>
#include <string.h>

#define NO_X_INCLUDES
#include "api.h"
#include "fonttool.h"
#include "options.h"

#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

typedef enum {
    FS_PROTO_ERROR,
    FS_PROTO_TCP,
    FS_PROTO_VMS,
    FS_PROTO_THIEF
} t_FontServerProtocol;

typedef struct {
    t_FontServerProtocol Protocol;
    CONST char *Name;
} t_FontServerProtocolName;

static t_FontServerProtocolName FontServerProtocolNames[] = {
    { FS_PROTO_ERROR, "error in font server string" },
    { FS_PROTO_TCP, "tcp" },
    { FS_PROTO_VMS, "decnet" },
    { FS_PROTO_THIEF, "font thief" },
    { 0, 0 }
};

int
parseOrderedList(start, endp, reportError, resultp)
    char *start;
    char **endp;
    CONST char **reportError;
    char ***resultp;
{
    int nCats = 1;
    register char *p;

    /* Take a string of comma-separated entries and make an array,
     * mallocing each one.
     *
     * Set endp to point to the first unhandled character -- \n or \0 usually.
     */
    if (!start) {
	if (reportError) {
	    *reportError = "internal error: null start in parseOrderedList";
	}
	if (endp) {
	    *endp = 0;
	}
	return -1;
    }

    if (!resultp) {
	if (reportError) {
	    *reportError = "internal error: null resultp in parseOrderedList";
	}
	if (endp) {
	    *endp = 0;
	}
	return -1;
    }

    /* count the number of entries */
    for (p = start; *p; p++) {
	if (*p == ',') {
	    ++nCats;
	    while (p[1] == ',') {
		p++; /* treat ,,, as a single comma */
	    }
	}
    }

    /* allocte the array to hold the result */
    *resultp = (char **) malloc(sizeof(char *) * (nCats + 1));
    if (!*resultp) {
	if (reportError) {
	    *reportError = "out of memory";
	}
	if (endp) {
	    *endp = p;
	}
	return -1;
    }

    nCats = 0;
    for (p = start; *p; p++) {
	register char *q;
	char *tmp;

	for (q = p; *q; q++) {
	    if (*q == ',') {
		break;
	    }
	}

	tmp = malloc((q - p) + 1);
	if (!tmp) {
	    if (reportError) {
		*reportError = "out of memory parsing catalogue list";
	    }
	    if (endp) {
		*endp = p;
	    }
	    if (resultp) {
		free((char *)resultp);
	    }
	    return -1;
	}
	(void) strncpy(tmp, p, ((q - p) + 1));
	tmp[q - p] = '\0';
	(*resultp)[nCats++] = tmp;

	while (*q == ',') {
	    q++;
	}
	p = q;
	if (!*p) {
	    break;
	}
	p--; /* so the ++ in the loop skips the comma */
    }

    (*resultp)[nCats] = (char *) 0;

    if (endp) {
	*endp = p;
    }
    return nCats;
}

t_FontServerProtocol
parseFontServerName(fontServerName, reportError, hostp, portp, catsp)
    char *fontServerName;
    CONST char **reportError;
    char **hostp;
    int *portp;
    char ***catsp;
{
    /* fontServerName is a name as it appears in
     * the font path --
     * e.g.
     * tcp/mieza:7100
     */
    char host[500];
    int port;
    register char *p;

    /* parse into host and port */
    if (strncmp(fontServerName, "tcp/", 4) != 0) {
	/* doesn't begin with tcp/ -> fail (I can't write the
	 * code for VMS, sorry)
	 *
	 */
	if (reportError) {
	    *reportError = "can only handle font servers starting in tcp/";
	}
	return FS_PROTO_ERROR;
    }

    for (p = &fontServerName[4]; *p; p++) {
	if (*p == ':') {
	    break;
	}
    }

    if (*p != ':') {
	if (reportError) {
	    *reportError = "font server name must contain a \":\"";
	}
	return FS_PROTO_ERROR;
    }

    if (p - fontServerName > sizeof(host)) {
	if (reportError) {
	    *reportError = "font server host name was too long";
	}
	return FS_PROTO_ERROR;
    }
    (void) strncpy(host, &fontServerName[4], (p - fontServerName) - 4);
    host[(p - fontServerName) - 4] = '\0';
    ++p;
    port = 0;

    if (!isdigit(*p)) {
	if (reportError) {
	    *reportError = "there was no port number after the \";\"";
	}
	return FS_PROTO_ERROR;
    }

    do {
	port *= 10;
	port += *p - '0';
	p++;
    } while (*p && isdigit(*p));

    if (!port) {
	if (reportError) {
	    *reportError = "zero is an implausible port number";
	}
	return FS_PROTO_ERROR;
    }

    if (*p != '\0' && *p != '/') {
	if (reportError) {
	    *reportError = "trailing characters found after port number";
	}
	return FS_PROTO_ERROR;
    }

    if (*p) {
	p++;
    }

    if (catsp) {
	if (!*p) {
	    *catsp = (char **) 0;
	} else {
	    char *end;
	    int nCats = parseOrderedList(p, &end, reportError, catsp);

	    if (nCats < 0) {
		/* reportError was set by parseOrderedList */
		return FS_PROTO_ERROR;
	    }
	}
    }

    if (portp) {
	*portp = port;
    }

    if (hostp) {
	*hostp = strdup(host);
    }

    return FS_PROTO_TCP;
}

/* config file syntax is
 * keyword = value
 * values are not continued over lines.
 */

typedef struct {

} t_fontServerConfigFile;

typedef enum {
    SKT_Error,
    SKT_ListOfString,
    SKT_String,
    SKT_Cardinal,
    SKT_Boolean,
    SKT_ListOfResolutions
} t_ServerKeywordType;

typedef struct {
    t_ServerKeywordType Type;
    CONST char *Name;
} t_ServerKeywordTypeName;

static t_ServerKeywordTypeName ServerKeywordTypeNames[] = {
    { SKT_ListOfString, "comma-separted list of strings" },
    { SKT_String, "string" },
    { SKT_Cardinal, "cardinal (whole number)" },
    { SKT_Boolean, "true or false" },
    { SKT_ListOfResolutions, "list of resolutions" },
    { SKT_Error, 0 }
};

typedef struct {
    CONST char *Name;
    t_ServerKeywordType Type;
    CONST char *description[8];
} t_ServerKeyword;

static t_ServerKeyword ServerKeywords[] = {
    {
	"catalogue",
	SKT_ListOfString,
	{
	    "Ordered list of font path element names.",
	    0,
	}
    },

    {
	"alternate-servers",
	SKT_ListOfString,
	{
	    "List of alternate servers for this font server.",
	    0,
	}
    },

    {
	"client-limit",
	SKT_Cardinal,
	{
	     "Number of clients this font server will support before",
	     "refusing service.",
	     "This is useful for tuning the load on each",
	     "individual font server.",
	    0,
	}
    },

    {
	"clone-self",
	SKT_Boolean,
	{
	    "Whether this font server should attempt to clone itself",
	    "when it reachs the client-limit.",
	    0,
	}
    },

    {
	"default-point-size",
	SKT_Cardinal,
	{
	    "The default pointsize (in decipoints) for fonts that",
	    "don't specify.  The default is 120.",
	    0,
	}
    },

    {
	"default-resolutions",
	SKT_ListOfResolutions,
	{
	    "Resolutions the server supports by default. This",
	    "information may be used as a hint for pre-rendering,",
	    "and substituted for scaled fonts which do not specify a",
	    "resolution.",
	    "A resolution is a comma-separated pair of",
	    "x and y resolutions in pixels per inch.",
	    "Multiple resolutions are separated by commas.",
	    0
	}
    },

    {
	"error-file",
	SKT_String,
	{
	    "Filename of the error file.",
	    "All warnings and errors will be logged here.",
	    0,
	}
    },

    {
	"port",
	SKT_Cardinal,
	{
	    "TCP port on which the server will listen for connections.",
	    0,
	}
    },

    {
	"use-syslog",
	SKT_Boolean,
	{
	    "Whether syslog(3) (on supported systems) is to be used for errors.",
	    0,
	}
    },

    {
	"deferglyphs",
	SKT_String,
	{
	    "Set the mode for delayed fetching and caching of",
	    "glyphs. Value is \"none\", meaning deferred glyphs is",
	    "disabled, \"all\", meaning it is enabled for all fonts,",
	    "and \"16\", meaning it is enabled only for 16-bits fonts.",
	    0,
	}
    },

    {
	0,
	0,
	{
	    0,
	}
    }
};

static void
OutputXMLFieldWithAtts(Name, Indent, Value, Atts)
    char *Name;
    int Indent;
    char *Value;
    char *Atts;
{
    if (!Value) {
	Value = (char *) "";
    }

    while (Indent-- > 0) {
	putchar(' ');
    }

    printf("<%s", Name);
    if (Atts && *Atts) {
	printf(" %s", Atts);
    }
    printf(">%s</%s>\n", Value, Name);
}

#define OutputXMLField(Name, Indent, Value) \
    OutputXMLFieldWithAtts(Name, Indent, Value, (char *) 0)

int
main(argc, argv)
    int argc;
    char *argv[];
{
    t_FontServerProtocol theProtocol;
    char *Host;
    int Port;
    CONST char *Error;
    char **catalogs;
    int exitStatus = 0;

    while (--argc > 0) {
	++argv;
	Error = 0;
	theProtocol = parseFontServerName(
	    argv[0], &Error, &Host, &Port, &catalogs
	);

	if (theProtocol == FS_PROTO_ERROR) {
	    if (Error) {
		printf("%s: error [%s]\n", argv[0], Error);
		if (exitStatus > 128) {
		    ++exitStatus;
		}
		continue;
	    }
	}

	printf("<fontServer>");
	{
	    OutputXMLField("fullName", 2, argv[0]);
	    OutputXMLField("protocol", 2, 
		FontServerProtocolNames[theProtocol].Name
	    );
	    OutputXMLField("host", 2, Host);
	    {
		char buf[100];
		(void) sprintf(buf, "%d", Port);
		OutputXMLField("port", 2, buf);
	    }
	    printf("  <catalogList>\n");
	    if (catalogs) {
		int i;

		for (i = 0; catalogs[i]; i++) {
		    OutputXMLField("catalog", 4, catalogs[i]);
		    free(catalogs[i]);
		    catalogs[i] = 0;
		}

	    } else {
		OutputXMLFieldWithAtts("catalog", 4, "all", "source=default");
	    }
	    printf("  </catalogList>\n");
	}
	printf("</fontServer>\n");
    }
    return exitStatus;
}
