#include <X11/X.h>
#include <X11/Xlib.h>

#include <stdlib.h>
#include <string.h>

#ifndef FILE
# include <stdio.h>
#endif

#include "fonttool.h"

static int sortOrder[] = {
    FIELD_FAMILY,
    FIELD_WEIGHT,
    FIELD_SLANT,
    FIELD_SETWIDTH,
    FIELD_ADDEDSTYLE,
    FIELD_FOUNDRY,
    FIELD_REGISTRY,
    FIELD_ENCODING,
    FIELD_POINTSIZE,
    FIELD_PIXELSZ,
    FIELD_RESX,
    FIELD_RESY,
    FIELD_SPACE,
    FIELD_AVERAGEWIDTH,
    -1
};

char *fieldNames[] = {
    (char *) "foundry",
    (char *) "family",
    (char *) "weight",
    (char *) "slant",
    (char *) "setwidth",
    (char *) "addedstyle",
    (char *) "pixelsz",
    (char *) "pointsize",
    (char *) "resx",
    (char *) "resy",
    (char *) "space",
    (char *) "averagewidth",
    (char *) "registry",
    (char *) "encoding",
    (char *) 0
};

int
fileNameToNumber(name)
    char *name;
{
    int i;

    for (i = 0; fieldNames[i]; i++) {
	if (strcmp(name, fieldNames[i]) == 0) {
	    return i;
	}
    }
    return -1;
}

#define ProgressMessage(m) /* nothing */
    /* We don't really want these messages, although they are useful
     * for debugging...
     */

#ifndef ProgressMessage

static int
ProgressMessage(message)
    char *message;
{
    /* fprintf(stderr, "@Progress: %s\n", message); */
}
#endif

static int
compareFonts(f1, f2)
    t_OneFont *f1, *f2;
{
    int f;

    for (f = 0; sortOrder[f] > 0; f++) {
	int i;
	i = strcmp(f1->Fields[sortOrder[f]], f2->Fields[sortOrder[f]]);
	if (i) return i;
    }
    return 0;
}

void
FontNameToFields(FontName, Fields)
    char *FontName;
    char **Fields;
{
    int f;
    register char *p;

    /* parse the fields */
    for (f = 0, p = FontName; *p; p++) {
	/* -*-*-*-*-*-*-*-*-*-*-*-*-*-*
	 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3
	 */
	if (*p == '-') {
	    *p = '\0';
	    Fields[f++] = &p[1];
	}
    }
}

void
FieldsToFontName(FontName, Fields)
    char *FontName;
    char **Fields;
{
    int f;
    register char *p;

    /* parse the fields */
    for (f = 0, p = FontName; f < 14; f++) {
	*p++ = '-';
	*p = '\0';
	if (Fields[f]) {
	    (void) strcat(p, Fields[f]);
	    while (*p) {
		p++;
	    }
	} else {
	    *p++ = '*';
	    *p = '\0';
	}
    }
}

void
destroyFonts(Fonts, n)
    t_OneFont *Fonts;
    int n;
{
    int i;

    for (i = 0; i < n; i++) {
	if (Fonts[n].FullName2) {
	    free(Fonts[n].FullName2);
	}
    }
    free((char *) Fonts);
}

int
GetFontNamesMatching(dpy, pattern, resultp)
    Display *dpy;
    char *pattern;
    t_OneFont **resultp;
{
    int maxnames = 99999;
    int actualcount;
    char **FontNames;
    int i;

    ProgressMessage("Asking server for the list of font names...");
    FontNames = XListFonts(dpy, pattern, maxnames, &actualcount);

    *resultp = (t_OneFont *) 0;

    if (!FontNames || actualcount <= 0) {
	ProgressMessage("No fonts matched");
	return 0;
    }

    ProgressMessage("Saving font names...");

    *resultp = (t_OneFont *) malloc(sizeof(t_OneFont) * actualcount);
    if (!*resultp) {
	ProgressMessage("Couldn't allocate enough memory.");
	XFreeFontNames(FontNames);
	return 0;
    }

    ProgressMessage("Parsing font names...");
    /* parse the individual font names */
    for (i = 0; i < actualcount; i++) {
	register t_OneFont *op;

	op = &(*resultp)[i];

	op->FullName = strdup(FontNames[i]);
	op->FullName2 = strdup(FontNames[i]);
	op->PostScriptName = (char *) "Courier"; /* TODO FIXME LOTS OF WORK! */
	
	/* parse the fields */
	FontNameToFields(op->FullName2, op->Fields);
    }

    /* save some server memory: */
    XFreeFontNames(FontNames);
    FontNames = 0; /* so we don't try and use it */

    /* sort the result */
    qsort(*resultp, actualcount, sizeof(t_OneFont), compareFonts);

    /* weed out duplicates */
    for (i = 0; i + 1 < actualcount; i++) {
	int j;
	int deleted;

	deleted = 0;
	for (j = i + 1; j < actualcount; j++) {
	    if (strcmp((*resultp)[i].FullName, (*resultp)[j].FullName) != 0) {
		break;
	    }
	    free((*resultp)[j].FullName);
	    free((*resultp)[j].FullName2);
	    ++deleted;
	}

	if (deleted) {
	    /* move them down; note: can't use memcpy here, because
	     * the regions of memory (src & dest) overlap
	     */
	    bcopy(
		&(*resultp)[j],
		&(*resultp)[i + 1], 
		(actualcount - (j - 1)) * sizeof(t_OneFont)
	    );
	    actualcount -= deleted;
	}
    }

    return actualcount;
}

#ifdef TEST_LISTFONTS
main(argc, argv)
    int argc;
    char *argv[];
{
    char *displayName = NULL;
    Display *dpy = XOpenDisplay(displayName);

    if (!dpy) {
	fprintf(stderr, "unable to connnect to display.");
	exit(1);
    }

    if (argc == 1) {
	printf("Usage: %s pattern [...]\n", argv[0]);
    }

    while (argc > 1) {
	t_OneFont *Fonts;
	int i;
	int FontCount;
	
	FontCount = GetFontNamesMatching(dpy, argv[1], &Fonts);

	for (i = 0; i < FontCount; i++) {
	    printf("%s: %s\n",
		Fonts[i].Fields[FIELD_FAMILY],
		Fonts[i].FullName
	    );
	}

	destroyFonts(Fonts);

	--argc;
    }

}
#endif /*TEST_LISTFONTS*/
